import { createRef } from 'react';
import { TestRenderer } from './TestRenderer.mjs';
import { HookRunner } from './HookRunner.mjs';
import { nodeName, toReactString } from './print.mjs';
import { jsx } from 'react/jsx-runtime';

const IS_NODE = Symbol.for('QuiltTesting.Node');
function createEnvironment(env) {
  const allRendered = new Set();
  const render = createRender({});
  return {
    render,
    createRender: createRender,
    rendered: allRendered,
    destroyAll
  };
  function createRoot(element, {
    context: rootContext,
    actions: rootActions,
    render,
    resolveRoot = defaultResolveRoot
  } = {}) {
    let rootNode = null;
    let mounted = false;
    let acting = false;
    let context;
    const abort = new AbortController();
    const testRendererRef = /*#__PURE__*/createRef();
    const rootApi = {
      act,
      mount,
      unmount,
      setProps,
      context: rootContext ?? {},
      actions: rootActions ?? {},
      signal: abort.signal
    };
    if (Symbol.dispose) {
      rootApi[Symbol.dispose] = () => {
        if (mounted) unmount();
      };
    }
    const root = new Proxy(rootApi, {
      get(target, key, receiver) {
        if (Reflect.ownKeys(target).includes(key)) {
          return Reflect.get(target, key, receiver);
        }
        return withRootNode(rootNode => Reflect.get(rootNode, key));
      }
    });
    return root;
    function createNode(createOptions) {
      const {
        type,
        props,
        instance,
        children
      } = createOptions;

      // We can’t just pick the remaining properties off `createOptions` with the
      // spread operator, because that attempts to invoke any getters in extensions
      // (including ones like .domNodes that may throw errors)
      const extensionPropertyDescriptors = Object.getOwnPropertyDescriptors(createOptions);
      delete extensionPropertyDescriptors.type;
      delete extensionPropertyDescriptors.props;
      delete extensionPropertyDescriptors.instance;
      delete extensionPropertyDescriptors.type;
      delete extensionPropertyDescriptors.children;
      let resolvedChildren;
      let resolvedDescendants;
      function getChildren() {
        if (resolvedChildren == null) {
          resolvedChildren = children();
        }
        return resolvedChildren;
      }
      function getDescendants() {
        if (resolvedDescendants == null) {
          resolvedDescendants = collectDescendants([], getChildren());
        }
        return resolvedDescendants;
      }
      const find = (type, props) => {
        return getDescendants().find(element => {
          return isNode(element) && isMatchingType(element.type, type) && (props == null || equalSubset(props, element.props));
        }) ?? null;
      };
      const baseNode = {
        [IS_NODE]: true,
        type,
        props,
        instance,
        get children() {
          return getChildren();
        },
        get descendants() {
          return getDescendants();
        },
        get text() {
          return getChildren().reduce((text, child) => `${text}${typeof child === 'string' ? child : child.text}`, '');
        },
        prop: key => props[key],
        is: checkType => isMatchingType(node.type, checkType),
        find,
        findAll: (type, props) => getDescendants().filter(element => isNode(element) && isMatchingType(element.type, type) && (props == null || equalSubset(props, element.props))),
        findWhere: predicate => getDescendants().find(element => isNode(element) && predicate(element)) ?? null,
        findAllWhere: predicate => getDescendants().filter(element => isNode(element) && predicate(element)),
        findContext: Context => find(Context.Provider)?.prop('value'),
        trigger: (prop, ...args) => act(() => {
          const propValue = props[prop];
          if (propValue == null) {
            throw new Error(`Attempted to call prop ${String(prop)} but it was not defined.`);
          }
          return propValue(...args);
        }, {
          eager: true
        }),
        triggerKeypath: (keypath, ...args) => act(() => {
          const parts = keypath.split(/[.[\]]/g).filter(Boolean);
          let currentProp = props;
          const currentKeypath = [];
          for (const part of parts) {
            if (currentProp == null || typeof currentProp !== 'object') {
              throw new Error(`Attempted to access field keypath '${currentKeypath.join('.')}', but it was not an PlainObject.`);
            }
            currentProp = currentProp[part];
            currentKeypath.push(part);
          }
          if (typeof currentProp !== 'function') {
            throw new Error(`Value at keypath '${keypath}' is not a function.`);
          }
          return currentProp(...args);
        }, {
          eager: true
        }),
        debug: options => toReactString(baseNode, options),
        toString: () => `<${nodeName(baseNode)} />`
      };
      const node = {};
      Object.defineProperties(node, {
        ...Object.getOwnPropertyDescriptors(baseNode),
        ...extensionPropertyDescriptors
      });
      return node;
    }
    function mount() {
      if (mounted) {
        throw new Error('Attempted to mount a node that was already mounted');
      }
      act(() => {
        context = env.mount( /*#__PURE__*/jsx(TestRenderer, {
          ref: testRendererRef,
          render: render,
          children: element
        }));
      });
      allRendered.add(root);
      mounted = true;
    }
    function unmount() {
      if (!mounted) {
        throw new Error('You attempted to unmount a node that was already unmounted');
      }
      assertRootNode();
      act(() => {
        env.unmount(context);
      });
      allRendered.delete(root);
      mounted = false;
      abort.abort();
    }
    function setProps(props) {
      assertRootNode();
      act(() => {
        testRendererRef.current?.setProps(props);
      });
    }
    function act(action, {
      update = true,
      eager = false
    } = {}) {
      const performUpdate = update ? updateRootNode : noop;
      let result;
      if (acting) {
        return action();
      }
      acting = true;
      const afterResolve = () => {
        acting = false;
        performUpdate();
        return result;
      };
      const actResult = env.act(() => {
        result = action();

        // The return type of non-async `act()`, DebugPromiseLike, contains a `then` method
        // This condition checks the returned value is an actual Promise and returns it
        // to React’s `act()` call, otherwise we just want to return `undefined`
        if (isPromise(result) && !eager) {
          return Promise.resolve(result).then(noop);
        }
      });
      if (isPromise(result)) {
        if (eager) {
          performUpdate();
          return act(() => Promise.resolve(result).then(noop)).then(afterResolve);
        } else {
          return Promise.resolve(actResult).then(afterResolve);
        }
      }
      return afterResolve();
    }
    function updateRootNode() {
      const testRenderer = testRendererRef.current;
      if (testRenderer == null) {
        rootNode = null;
        return;
      }
      rootNode = env.update(testRenderer, createNode, context);
      rootNode = rootNode && resolveRoot(rootNode);
      const error = testRenderer.getError();
      if (error) throw error;
    }
    function withRootNode(perform) {
      assertRootNode();
      return perform(rootNode);
    }
    function assertRootNode() {
      if (rootNode == null) {
        throw new Error('Attempted to operate on a mounted tree, but the component is no longer mounted');
      }
    }
  }
  function defaultResolveRoot(node) {
    return typeof node.children[0] === 'string' ? node : node.children[0];
  }
  function destroyAll() {
    for (const wrapper of allRendered) {
      wrapper.unmount();
    }
  }
  function createRender(createRenderOptions) {
    const {
      options: customizeOptions,
      render = defaultRender,
      context: createContext = defaultContext,
      actions: createActions = defaultActions,
      afterRender
    } = createRenderOptions;
    function mount(element, options = {}) {
      const resolvedOptions = customizeOptions ? {
        ...options,
        ...customizeOptions(options)
      } : options;
      const context = createContext(resolvedOptions, {});
      const actions = {};
      const root = createRoot(element, {
        context,
        actions,
        render: element => render(element, context, resolvedOptions),
        resolveRoot: root => root.find(element.type)
      });
      Object.assign(actions, createActions(root, resolvedOptions, {}));
      root.mount();
      if (afterRender) {
        const afterRenderResult = root.act(() => afterRender(root, resolvedOptions));
        return isPromise(afterRenderResult) ? afterRenderResult.then(() => root) : root;
      } else {
        return root;
      }
    }
    function testHook(useHook, options) {
      const hookRunnerRef = /*#__PURE__*/createRef();
      const rootOrPromise = mount( /*#__PURE__*/jsx(HookRunner, {
        useHook: useHook,
        ref: hookRunnerRef
      }), options);
      const withRoot = root => {
        const getCurrentValue = () => {
          if (hookRunnerRef.current == null) {
            throw new Error('Attempted to access the hook value while the hook runner was not mounted');
          }
          return hookRunnerRef.current.current;
        };
        return {
          get current() {
            return getCurrentValue();
          },
          get value() {
            return getCurrentValue();
          },
          get context() {
            return root.context;
          },
          get actions() {
            return root.actions;
          },
          act(action) {
            return root.act(() => action(getCurrentValue()));
          },
          mount() {
            root.mount();
          },
          unmount() {
            root.unmount();
          }
        };
      };
      return 'then' in rootOrPromise ? rootOrPromise.then(root => withRoot(root)) : withRoot(rootOrPromise);
    }
    Reflect.defineProperty(mount, 'hook', {
      writable: false,
      value: testHook
    });
    Reflect.defineProperty(mount, 'extend', {
      writable: false,
      value: ({
        options: customizeOptions,
        context: createAdditionalContext = defaultContext,
        actions: createAdditionalActions = defaultActions,
        render: additionalRender = defaultRender,
        afterRender: additionalAfterRender = noop
      }) => {
        return createRender({
          options: customizeOptions,
          context(options, initialContext) {
            const baseContext = {
              ...initialContext,
              ...createContext(options, initialContext)
            };
            const additionalContext = createAdditionalContext(options, baseContext);
            return {
              ...baseContext,
              ...additionalContext
            };
          },
          actions(root, options, initialActions) {
            const baseActions = {
              ...initialActions,
              ...createActions(root, options, initialActions)
            };
            const additionalActions = createAdditionalActions(root, options, baseActions);
            return {
              ...baseActions,
              ...additionalActions
            };
          },
          render: (element, context, options) => render(additionalRender(element, context, options), context, options),
          afterRender: (wrapper, options) => {
            const result = additionalAfterRender(wrapper, options);
            const finalResult = () => afterRender ? afterRender(wrapper, options) : undefined;
            return isPromise(result) ? result.then(finalResult) : finalResult();
          }
        });
      }
    });
    return mount;
  }
}
function defaultRender(element) {
  return element;
}
function defaultContext() {
  return {};
}
function defaultActions() {
  return {};
}
function noop() {}
function isNode(maybeNode) {
  return maybeNode != null && maybeNode[IS_NODE];
}

// We do a highly optimized loop since this can be called for very
// large trees of components.
function collectDescendants(descendants, children) {
  for (let i = 0; i < children.length; i++) {
    const child = children[i];
    descendants.push(child);
    if (typeof child !== 'string' && child.children.length > 0) {
      collectDescendants(descendants, child.children);
    }
  }
  return descendants;
}
function isPromise(promise) {
  return typeof promise?.then === 'function';
}
function isMatchingType(type, test) {
  if (type === test) {
    return true;
  }
  if (test == null) {
    return false;
  }
  return test.type != null && isMatchingType(type, test.type);
}
function equalSubset(subset, full) {
  return Object.keys(subset).every(key => key in full && full[key] === subset[key]);
}

export { createEnvironment, isNode };
